@@ -41,7 +41,8 @@
  * @inlen: length of input array with unicode code points.
  * @out: output zero terminated string that must have room for at
  *       least 63 characters plus the terminating zero.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * The ToASCII operation takes a sequence of Unicode code points that make
  * up one label and transforms it into a sequence of code points in the
@@ -63,7 +64,7 @@
  * operation multiple times has exactly the same effect as applying it just
  * once.
  *
- * Return value: Returns 0 on success, or an error code.
+ * Return value: Returns 0 on success, or an #Idna_rc error code.
  */
 int
 idna_to_ascii_4i (const uint32_t * in, size_t inlen, char *out, int flags)
@@ -115,10 +116,16 @@ idna_to_ascii_4i (const uint32_t * in, size_t inlen, char *out, int flags)
     len = strlen (p);
     do
       {
+	char *newp;
+
 	len = 2 * len + 10;	/* XXX better guess? */
-	p = realloc (p, len);
-	if (p == NULL)
-	  return IDNA_MALLOC_ERROR;
+	newp = realloc (p, len);
+	if (newp == NULL)
+	  {
+	    free (p);
+	    return IDNA_MALLOC_ERROR;
+	  }
+	p = newp;
 
 	if (flags & IDNA_ALLOW_UNASSIGNED)
 	  rc = stringprep_nameprep (p, len);
@@ -251,7 +258,7 @@ step8:
   return IDNA_SUCCESS;
 }
 
-/* ToUnicode().  May realloc() utf8in. */
+/* ToUnicode().  May realloc() utf8in.  Will free utf8in unconditionally. */
 static int
 idna_to_unicode_internal (char *utf8in,
 			  uint32_t * out, size_t * outlen, int flags)
@@ -288,9 +295,13 @@ idna_to_unicode_internal (char *utf8in,
    */
   do
     {
-      utf8in = realloc (utf8in, utf8len + addlen);
-      if (!utf8in)
-	return IDNA_MALLOC_ERROR;
+      char *newp = realloc (utf8in, utf8len + addlen);
+      if (newp == NULL)
+	{
+	  free (utf8in);
+	  return IDNA_MALLOC_ERROR;
+	}
+      utf8in = newp;
       if (flags & IDNA_ALLOW_UNASSIGNED)
 	rc = stringprep_nameprep (utf8in, utf8len + addlen);
       else
@@ -300,7 +311,10 @@ idna_to_unicode_internal (char *utf8in,
   while (rc == STRINGPREP_TOO_SMALL_BUFFER);
 
   if (rc != STRINGPREP_OK)
-    return IDNA_STRINGPREP_ERROR;
+    {
+      free (utf8in);
+      return IDNA_STRINGPREP_ERROR;
+    }
 
   /* 3. Verify that the sequence begins with the ACE prefix, and save a
    * copy of the sequence.
@@ -308,7 +322,10 @@ idna_to_unicode_internal (char *utf8in,
 
 step3:
   if (memcmp (IDNA_ACE_PREFIX, utf8in, strlen (IDNA_ACE_PREFIX)) != 0)
-    return IDNA_NO_ACE_PREFIX;
+    {
+      free (utf8in);
+      return IDNA_NO_ACE_PREFIX;
+    }
 
   /* 4. Remove the ACE prefix.
    */
@@ -325,7 +342,10 @@ step3:
 
   rc = punycode_decode (strlen (utf8in), utf8in, outlen, out, NULL);
   if (rc != PUNYCODE_SUCCESS)
-    return IDNA_PUNYCODE_ERROR;
+    {
+      free (utf8in);
+      return IDNA_PUNYCODE_ERROR;
+    }
 
   out[*outlen] = 0;		/* add zero */
 
@@ -334,18 +354,25 @@ step3:
 
   rc = idna_to_ascii_4i (out, *outlen, tmpout, flags);
   if (rc != IDNA_SUCCESS)
-    return rc;
+    {
+      free (utf8in);
+      return rc;
+    }
 
   /* 7. Verify that the result of step 6 matches the saved copy from
    * step 3, using a case-insensitive ASCII comparison.
    */
 
   if (strcasecmp (utf8in, tmpout + strlen (IDNA_ACE_PREFIX)) != 0)
-    return IDNA_ROUNDTRIP_VERIFY_ERROR;
+    {
+      free (utf8in);
+      return IDNA_ROUNDTRIP_VERIFY_ERROR;
+    }
 
   /* 8. Return the saved copy from step 5.
    */
 
+  free (utf8in);
   return IDNA_SUCCESS;
 }
 
@@ -356,7 +383,8 @@ step3:
  * @out: output array with unicode code points.
  * @outlen: on input, maximum size of output array with unicode code points,
  *          on exit, actual size of output array with unicode code points.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * The ToUnicode operation takes a sequence of Unicode code points
  * that make up one label and returns a sequence of Unicode code
@@ -376,13 +404,12 @@ step3:
  * AllowUnassigned flag, and the UseSTD3ASCIIRules flag. The output of
  * ToUnicode is always a sequence of Unicode code points.
  *
- * Return value: Returns error condition, but it must only be used for
- *               debugging purposes.  The output buffer is always
- *               guaranteed to contain the correct data according to
- *               the specification (sans malloc induced errors).  NB!
- *               This means that you normally ignore the return code
- *               from this function, as checking it means breaking the
- *               standard.
+ * Return value: Returns #Idna_rc error condition, but it must only be
+ *   used for debugging purposes.  The output buffer is always
+ *   guaranteed to contain the correct data according to the
+ *   specification (sans malloc induced errors).  NB!  This means that
+ *   you normally ignore the return code from this function, as
+ *   checking it means breaking the standard.
  */
 int
 idna_to_unicode_44i (const uint32_t * in, size_t inlen,
@@ -404,7 +431,7 @@ idna_to_unicode_44i (const uint32_t * in, size_t inlen,
       *outlen = inlen;
     }
 
-  free (p);
+  /* p is freed in idna_to_unicode_internal.  */
 
   return rc;
 }
@@ -415,7 +442,8 @@ idna_to_unicode_44i (const uint32_t * in, size_t inlen,
  * idna_to_ascii_4z:
  * @input: zero terminated input Unicode string.
  * @output: pointer to newly allocated output string.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert UCS-4 domain name to ASCII string.  The domain name may
  * contain several labels, separated by dots.  The output buffer must
@@ -479,9 +507,13 @@ idna_to_ascii_4z (const uint32_t * input, char **output, int flags)
 
       if (out)
 	{
-	  out = realloc (out, strlen (out) + 1 + strlen (buf) + 1);
-	  if (!out)
-	    return IDNA_MALLOC_ERROR;
+	  char *newp = realloc (out, strlen (out) + 1 + strlen (buf) + 1);
+	  if (!newp)
+	    {
+	      free (out);
+	      return IDNA_MALLOC_ERROR;
+	    }
+	  out = newp;
 	  strcat (out, ".");
 	  strcat (out, buf);
 	}
@@ -506,7 +538,8 @@ idna_to_ascii_4z (const uint32_t * input, char **output, int flags)
  * idna_to_ascii_8z:
  * @input: zero terminated input UTF-8 string.
  * @output: pointer to newly allocated output string.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert UTF-8 domain name to ASCII string.  The domain name may
  * contain several labels, separated by dots.  The output buffer must
@@ -535,9 +568,11 @@ idna_to_ascii_8z (const char *input, char **output, int flags)
 
 /**
  * idna_to_ascii_lz:
- * @input: zero terminated input UTF-8 string.
+ * @input: zero terminated input string encoded in the current locale's
+ *   character set.
  * @output: pointer to newly allocated output string.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert domain name in the locale's encoding to ASCII string.  The
  * domain name may contain several labels, separated by dots.  The
@@ -566,7 +601,8 @@ idna_to_ascii_lz (const char *input, char **output, int flags)
  * idna_to_unicode_4z4z:
  * @input: zero-terminated Unicode string.
  * @output: pointer to newly allocated output Unicode string.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert possibly ACE encoded domain name in UCS-4 format into a
  * UCS-4 string.  The domain name may contain several labels,
@@ -605,9 +641,16 @@ idna_to_unicode_4z4z (const uint32_t * input, uint32_t ** output, int flags)
 
       if (out)
 	{
-	  out = realloc (out, sizeof (out[0]) * (outlen + 1 + buflen + 1));
-	  if (!out)
-	    return IDNA_MALLOC_ERROR;
+	  uint32_t *newp = realloc (out,
+				    sizeof (out[0])
+				    * (outlen + 1 + buflen + 1));
+	  if (!newp)
+	    {
+	      free (buf);
+	      free (out);
+	      return IDNA_MALLOC_ERROR;
+	    }
+	  out = newp;
 	  out[outlen++] = 0x002E;	/* '.' (full stop) */
 	  memcpy (out + outlen, buf, sizeof (buf[0]) * buflen);
 	  outlen += buflen;
@@ -634,7 +677,8 @@ idna_to_unicode_4z4z (const uint32_t * input, uint32_t ** output, int flags)
  * idna_to_unicode_8z4z:
  * @input: zero-terminated UTF-8 string.
  * @output: pointer to newly allocated output Unicode string.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert possibly ACE encoded domain name in UTF-8 format into a
  * UCS-4 string.  The domain name may contain several labels,
@@ -664,7 +708,8 @@ idna_to_unicode_8z4z (const char *input, uint32_t ** output, int flags)
  * idna_to_unicode_8z8z:
  * @input: zero-terminated UTF-8 string.
  * @output: pointer to newly allocated output UTF-8 string.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert possibly ACE encoded domain name in UTF-8 format into a
  * UTF-8 string.  The domain name may contain several labels,
@@ -694,7 +739,8 @@ idna_to_unicode_8z8z (const char *input, char **output, int flags)
  * @input: zero-terminated UTF-8 string.
  * @output: pointer to newly allocated output string encoded in the
  *   current locale's character set.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert possibly ACE encoded domain name in UTF-8 format into a
  * string encoded in the current locale's character set.  The domain
@@ -725,7 +771,8 @@ idna_to_unicode_8zlz (const char *input, char **output, int flags)
  *   character set.
  * @output: pointer to newly allocated output string encoded in the
  *   current locale's character set.
- * @flags: IDNA flags, e.g. IDNA_ALLOW_UNASSIGNED or IDNA_USE_STD3_ASCII_RULES.
+ * @flags: an #Idna_flags value, e.g., %IDNA_ALLOW_UNASSIGNED or
+ *   %IDNA_USE_STD3_ASCII_RULES.
  *
  * Convert possibly ACE encoded domain name in the locale's character
  * set into a string encoded in the current locale's character set.
